"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.associate_dhcp_options
hub.exec.boto3.client.ec2.create_dhcp_options
hub.exec.boto3.client.ec2.delete_dhcp_options
hub.exec.boto3.client.ec2.describe_dhcp_options
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "absent": {"require": ["aws.ec2.vpc.absent"]},
}


async def present(
    hub,
    ctx,
    name: str,
    dhcp_configurations: List[
        make_dataclass(
            "NewDhcpConfiguration",
            [
                ("Key", str, field(default=None)),
                ("Values", List[str], field(default=None)),
            ],
        )
    ] = None,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    vpc_id: List[str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a set of DHCP options for your VPC. After creating the set, you must associate it with the VPC, causing
    all existing and new instances that you launch in the VPC to use this set of DHCP options. The following are the
    individual DHCP options you can specify. For more information about the options, see RFC 2132.    domain-name-
    servers - The IP addresses of up to four domain name servers, or AmazonProvidedDNS. The default DHCP option set
    specifies AmazonProvidedDNS. If specifying more than one domain name server, specify the IP addresses in a
    single parameter, separated by commas. To have your instance receive a custom DNS hostname as specified in
    domain-name, you must set domain-name-servers to a custom DNS server.    domain-name - If you're using
    AmazonProvidedDNS in us-east-1, specify ec2.internal. If you're using AmazonProvidedDNS in another Region,
    specify region.compute.internal (for example, ap-northeast-1.compute.internal). Otherwise, specify a domain name
    (for example, ExampleCompany.com). This value is used to complete unqualified DNS hostnames. Important: Some
    Linux operating systems accept multiple domain names separated by spaces. However, Windows and other Linux
    operating systems treat the value as a single domain, which results in unexpected behavior. If your DHCP options
    set is associated with a VPC that has instances with multiple operating systems, specify only one domain name.
    ntp-servers - The IP addresses of up to four Network Time Protocol (NTP) servers.    netbios-name-servers - The
    IP addresses of up to four NetBIOS name servers.    netbios-node-type - The NetBIOS node type (1, 2, 4, or 8).
    We recommend that you specify 2 (broadcast and multicast are not currently supported). For more information
    about these node types, see RFC 2132.   Your VPC automatically starts out with a set of DHCP options that
    includes only a DNS server that we provide (AmazonProvidedDNS). If you create a set of options, and if your VPC
    has an internet gateway, make sure to set the domain-name-servers option either to AmazonProvidedDNS or to a
    domain name server of your choice. For more information, see DHCP options sets in the Amazon Virtual Private
    Cloud User Guide.

    Args:
        name(Text): An Idem name to identify the dhcp option resource.
        dhcp_configurations (List): A DHCP configuration option.
            * Key (str, optional): The name of a DHCP option. Options are domain-name-servers|domain-name|ntp-servers|netbios-name-servers|netbios-node-type
            * Values (List[str], optional): One or more values for the DHCP option.
                * Value (string): The attribute value. The value is case-sensitive.
        resource_id(Text, optional): AWS DHCP Option Set ID.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the dhcp option resource. Defaults to None.
            * Key (string) --  The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.
        vpc_id (List, optional): The list of AWS VPC ID that needs to be associated with dhcp options set.

    Request Syntax:
        [dhcp_option-name]:
          aws.ec2.dhcp_option.present:
          - resource_id: 'string'
          - dhcp_configurations:
              - Key: 'string'
                Values: List
                    - Value: 'string'
          - vpc_id: 'List'
          - tags:
              - Key: 'string'
                Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-dhcp-option:
              aws.ec2.dhcp_option.present:
                - dhcp_configurations:
                    - Key: domain-name-servers
                      Values:
                            - Value : 10.2.5.1
                            - Value : 10.2.5.2
                - vpc_id:
                  - vpc-76db75b8
                - tags:
                    - Key: test_name
                      Value: test-dhcp
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "ec2", "DhcpOptions", resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        convert_ret = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.ec2.dhcp_option '{name}' already exists",)
        if vpc_id is not None:
            if not ctx.get("test", False):
                vpc_ret = await hub.exec.boto3.client.ec2.describe_vpcs(
                    ctx, VpcIds=vpc_id
                )
                if not vpc_ret["result"]:
                    result["comment"] = vpc_ret["comment"]
                    result["result"] = vpc_ret["result"]
                for vpc_associate in vpc_id:
                    for vpc_dhcp in vpc_ret["ret"]["Vpcs"]:
                        if (
                            vpc_ret["result"]
                            and vpc_dhcp["DhcpOptionsId"] != resource_id
                        ):
                            update_ret = (
                                await hub.exec.boto3.client.ec2.associate_dhcp_options(
                                    ctx,
                                    DhcpOptionsId=resource_id,
                                    VpcId=vpc_associate,
                                )
                            )
                            result["comment"] = update_ret["comment"]
                            result["result"] = update_ret["result"]
                            resource_updated = resource_updated or bool(
                                update_ret["ret"]
                            )
            else:
                plan_state["vpc_id"] = vpc_id

        # check if tags need to be updated
        if tags is not None and tags != result["old_state"].get("tags"):
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=result["old_state"].get("resource_id"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["result"])
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "dhcp_configurations": dhcp_configurations,
                    "vpc_id": vpc_id,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.dhcp_option", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.create_dhcp_options(
            ctx,
            DhcpConfigurations=dhcp_configurations,
            TagSpecifications=[
                {
                    "ResourceType": "dhcp-options",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags is not None
            else None,
        )
        resource_id = ret["ret"]["DhcpOptions"]["DhcpOptionsId"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.dhcp_option", name=name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        if result["result"] and vpc_id is not None:
            for vpc_associate in vpc_id:
                update_ret = await hub.exec.boto3.client.ec2.associate_dhcp_options(
                    ctx,
                    DhcpOptionsId=ret["ret"]["DhcpOptions"]["DhcpOptionsId"],
                    VpcId=vpc_associate,
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if not result["result"]:
                    result["comment"] = update_ret["comment"]
                    return result
    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        resource = hub.tool.boto3.resource.create(
            ctx, "ec2", "DhcpOptions", resource_id
        )
        after = await hub.tool.boto3.resource.describe(resource)
        convert_ret = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_to_present(
                ctx, raw_resource=after, idem_resource_name=name
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["new_state"] = convert_ret["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified set of DHCP options. You must disassociate the set of DHCP options before you can delete
    it. You can disassociate the set of DHCP options by associating either a new set of options or the default set
    of options with the VPC.

    Args:
        name(Text): An Idem name to identify the dhcp option resource.
        resource_id(Text, optional): AWS DHCP Option Set ID. Idem automatically considers this resource being
         absent if this field is not specified.

    Request Syntax:
        [dhcp_option-name]:
          aws.ec2.dhcp_option.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.dhcp_option.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.dhcp_option", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "DhcpOptions", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.dhcp_option", name=name
        )
        return result
    else:
        convert_ret = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.dhcp_option", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.delete_dhcp_options(
            ctx, DhcpOptionsId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.dhcp_option", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your DHCP options sets. For more information, see DHCP options sets in the Amazon
    Virtual Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.dhcp_option
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_dhcp_options(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe dhcp_options {ret['comment']}")
        return result

    for resource in ret["ret"]["DhcpOptions"]:
        resource_id = resource.get("DhcpOptionsId")
        convert_ret = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_to_present(
                ctx, raw_resource=resource, idem_resource_name=resource_id
            )
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe dhcp options '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        translated_resource = convert_ret["ret"]
        result[translated_resource["resource_id"]] = {
            "aws.ec2.dhcp_option.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Exec functions for AWS API Gateway v2 Integration resources.
"""


async def update(
    hub,
    ctx,
    api_id: str,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 Integration resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Integration resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "ConnectionId": "connection_id",
            "ConnectionTypes": "connection_type",
            "ContentHandlingStrategy": "content_handling_strategy",
            "CredentialsArn": "credentials_arn",
            "Description": "description",
            "IntegrationMethod": "integration_method",
            "IntegrationSubtype": "integration_subtype",
            "IntegrationType": "integration_type",
            "IntegrationUri": "integration_uri",
            "PassthroughBehavior": "passthrough_behavior",
            "PayloadFormatVersion": "payload_format_version",
            "RequestParameters": "request_parameters",
            "RequestTemplates": "request_templates",
            "ResponseParameters": "response_parameters",
            "TemplateSelectionExpression": "template_selection_expression",
            "TimeoutInMillis": "timeout_in_millis",
            "TlsConfig": "tls_config",
        }
    )

    parameters_to_update = {}

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_integration(
                ctx,
                ApiId=api_id,
                IntegrationId=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result



import copy
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def update_user_tags(
    hub,
    ctx,
    user_name: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Update tags of AWS IAM user

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        user_name: AWS IAM username
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            tag_keys = list(tags_to_remove.keys())
            delete_ret = await hub.exec.boto3.client.iam.untag_user(
                ctx, UserName=user_name, TagKeys=tag_keys
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_user(
                ctx,
                UserName=user_name,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result


async def update_user(
    hub,
    ctx,
    before: Dict[str, Any],
    user_name: str = None,
    path: str = None,
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    if user_name and before.get("user_name") != user_name:
        update_payload["NewUserName"] = user_name
    if path and before.get("path") != path:
        update_payload["NewPath"] = path
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.iam.update_user(
                ctx=ctx, UserName=before.get("resource_id"), **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        if "NewUserName" in update_payload:
            result["ret"]["user_name"] = update_payload["NewUserName"]
            result["comment"] = result["comment"] + (
                f"Update user_name: {update_payload['NewUserName']} on user {before.get('user_name')}",
            )
        if "NewPath" in update_payload:
            result["ret"]["path"] = update_payload["NewPath"]
            result["comment"] = result["comment"] + (
                f"Update path: {update_payload['NewPath']} on user {before.get('user_name')}",
            )
    return result


async def list_(hub, ctx):
    ret = await hub.exec.boto3.client.iam.list_users(ctx)

    if not ret["result"]:
        return {
            "result": False,
            "ret": [],
            "comment": f"Can not list users: {ret['comment']}",
        }

    users = []
    for user_raw in ret["ret"]["Users"]:
        # we're using this function for the camel case to snake case
        user = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(user_raw)
        users.append(user)

    return {"result": True, "ret": users}



#!/usr/bin/env python3
import os
import pathlib
import shutil

from setuptools import Command
from setuptools import setup

NAME = "idem_aws"
DESC = ""

# Version info -- read without importing
_locals = {}
with pathlib.Path(NAME, "version.py").open() as fp:
    exec(fp.read(), None, _locals)

VERSION = _locals["version"]
SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()

with open("README.rst", encoding="utf-8") as f:
    LONG_DESC = f.read()

with pathlib.Path("requirements", "base.txt").open("r") as f:
    REQUIREMENTS = f.read().splitlines()

REQUIREMENTS_EXTRA = {}
EXTRA_PATH = pathlib.Path("requirements", "extra")
if EXTRA_PATH.exists():
    REQUIREMENTS_EXTRA["full"] = set()
    for extra in EXTRA_PATH.iterdir():
        with extra.open("r") as f:
            REQUIREMENTS_EXTRA[extra.stem] = f.read().splitlines()
            REQUIREMENTS_EXTRA["full"].update(REQUIREMENTS_EXTRA[extra.stem])


class Clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in (NAME, "tests"):
            for root, dirs, files in os.walk(
                os.path.join(os.path.dirname(__file__), subdir)
            ):
                for dir_ in dirs:
                    if dir_ == "__pycache__":
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for package in (NAME,):
        for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, package)):
            pdir = os.path.relpath(root, SETUP_DIRNAME)
            if "{{cookiecutter.root_dir}}" in pdir or "aws_auto" in pdir:
                continue
            modname = pdir.replace(os.sep, ".")
            modules.append(modname)
    return modules


setup(
    name="idem-aws",
    author="VMware Inc.",
    author_email="tyjohnson@vmware.com",
    url="https://gitlab.com/saltstack/pop/idem-aws",
    version=VERSION,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRA,
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    python_requires=">=3.7",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
    ],
    packages=discover_packages(),
    cmdclass={"clean": Clean},
)



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS Dynamo DB resources

    Args:
        resource_arn: Identifies the Amazon DynamoDB resource to which tags should be added.
            This value is an Amazon Resource Name (ARN).
        old_tags: Dict of existing tags
        new_tags: Dict of new tags

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": dict of updated tags}

    """

    tags_to_add = {}
    tags_to_remove = []
    for key, value in new_tags.items():
        if key in old_tags and old_tags[key] != value or (key not in old_tags):
            tags_to_add[key] = value
    for key in old_tags:
        if key not in new_tags:
            tags_to_remove.append(key)
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_error_message = f"Could not delete tags {tags_to_remove} for DynamoDB table with ARN {resource_arn}"
            delete_ret = await hub.exec.boto3.client.dynamodb.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                hub.log.debug(delete_error_message)
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_error_message = f"Could not create tags {tags_to_remove} for DynamoDB table with ARN {resource_arn}"
            create_tag_resp = await hub.exec.boto3.client.dynamodb.tag_resource(
                ctx,
                ResourceArn=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not create_tag_resp:
                hub.log.debug(add_error_message)
                result["comment"] = result["comment"] + (
                    f"Could not update tags {tags_to_add}",
                )
                result["result"] = False
                return result

    result["ret"] = new_tags
    result["comment"] = (
        f"Updated tags: Added [{tags_to_add}] Removed [{tags_to_remove}]",
    )
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.s3.bucket.present"]},
}

BUCKET_NOT_EXISTS_ERROR_MESSAGE = "ClientError: An error occurred (404) when calling the HeadBucket operation: Not Found"

STATE_NAME = "aws.s3.bucket_notification"


async def present(
    hub,
    ctx,
    name: str,
    notifications: List[
        make_dataclass(
            "LambdaFunctionConfiguration",
            [
                ("LambdaFunctionArn", str),
                ("Events", List[str]),
                ("Id", str, field(default=None)),
                (
                    "Filter",
                    make_dataclass(
                        "NotificationConfigurationFilter",
                        [
                            (
                                "Key",
                                make_dataclass(
                                    "S3KeyFilter",
                                    [
                                        (
                                            "FilterRules",
                                            List[
                                                make_dataclass(
                                                    "FilterRuleList",
                                                    [
                                                        (
                                                            "Name",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "Value",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                    ],
                                                )
                                            ],
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            )
                        ],
                    ),
                    field(default=None),
                ),
            ],
        )
    ],
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Bucket Notification allows user to receive notifications when certain events happen in the S3 bucket. To enable
    notifications, add a notification configuration that identifies the events that you want Amazon S3 to publish.
    Make sure that it also identifies the destinations where you want Amazon S3 to send the notifications. You store
    this configuration in the notification subresource that's associated with a bucket
    Amazon S3 can publish notifications for the following events:

    a) New object created events
    b) Object removal events
    c) Restore object events
    d) Reduced Redundancy Storage (RRS) object lost events
    e) Replication events
    f) S3 Lifecycle expiration events
    g) S3 Lifecycle transition events
    h) S3 Intelligent-Tiering automatic archival events
    i) Object tagging events
    j) Object ACL PUT events

    Amazon S3 can send event notification messages to the following destinations. User can specify the Amazon Resource
    Name (ARN) value of these destinations in the notification configuration.

    1) Amazon Simple Notification Service (Amazon SNS) topics
    2) Amazon Simple Queue Service (Amazon SQS) queues
    3) AWS Lambda function

    Args:
        name(Text): Name of the bucket on which notification needs to be configured.
        notifications(List[Dict[str, Any]], optional): Describes the Lambda functions to invoke and the events for which to invoke them.
            * Id (str, optional): An optional unique identifier for configurations in a notification configuration. If you don't
                provide one, Amazon S3 will assign an ID.
            * LambdaFunctionArn (str): The Amazon Resource Name (ARN) of the Lambda function that Amazon S3 invokes when the specified
                event type occurs.
            * Events (List[str]): The Amazon S3 bucket event for which to invoke the Lambda function. For more information, see
                Supported Event Types in the Amazon S3 User Guide.
            * Filter (Dict[str, Any], optional): Specifies object key name filtering rules. For information about key name filtering, see
                Configuring Event Notifications in the Amazon S3 User Guide.
                * Key (Dict[str, Any], optional): A container for object key name prefix and suffix filtering rules.
                    * FilterRules (List[Dict[str, Any]], optional): A list of containers for the key-value pair that defines the criteria for the filter rule.
                        * Name (str, optional): The object key name prefix or suffix identifying one or more objects to which the filtering rule
                            applies. The maximum length is 1,024 characters. Overlapping prefixes and suffixes are not
                            supported. For more information, see Configuring Event Notifications in the Amazon S3 User
                            Guide.
                        * Value (str, optional): The value that the filter searches for in object key names.
        resource_id(Text, optional): Name of the bucket and 'notifications' keyword with a seperator '-'.

    Returns:
        Dict[str, Any]

    Request Syntax:
        [bucket_name]-notifications:
              aws.s3.bucket_notification.present:
                - name: [string]
                - resource_id: [string]
                - notifications
                    LambdaFunctionConfigurations:
                        - Events:
                          - [string]
                          Filter:
                            Key:
                              FilterRules:
                              - Name: Prefix
                                Value: ''
                              - Name: Suffix
                                Value: [string]
                          Id: [string]
                          LambdaFunctionArn: [string]
                    TopicConfigurations:

    Examples:

        .. code-block:: sls

            test-bucket-notifications:
              aws.s3.bucket_notification.present:
                - name: test-bucket
                - resource_id: test-bucket-notifications
                - notifications
                    LambdaFunctionConfigurations:
                        - Events:
                          - s3:ObjectCreated:*
                          Filter:
                            Key:
                              FilterRules:
                              - Name: Prefix
                                Value: ''
                              - Name: Suffix
                                Value: .jpeg
                          Id: test
                          LambdaFunctionArn: arn:aws:lambda:us-west-1:000000000000:function:test
                    TopicConfigurations:
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if resource_id:
        name = resource_id.split("-notifications")[0]

    # Check if bucket exists
    ret = await hub.exec.boto3.client.s3.head_bucket(ctx, Bucket=name)
    if not ret["result"]:
        if BUCKET_NOT_EXISTS_ERROR_MESSAGE in ret["comment"]:
            result["comment"] = (f"aws.s3.bucket with name: {name} does not exists",)
        else:
            result["comment"] = ret["comment"]
        result["result"] = False
        return result

    # Get current notification configurations for the bucket
    bucket_notification_response = (
        await hub.exec.boto3.client.s3.get_bucket_notification_configuration(
            ctx, Bucket=name
        )
    )

    if bucket_notification_response["result"]:
        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_notification_to_present(
            bucket_notification_response["ret"], name
        )
    else:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    # check and create payload for creating notification configurations to avoid redundant calls
    notifications_diff = hub.tool.aws.s3.s3_utils.get_notification_configuration_diff(
        notifications, result["old_state"]
    )

    if notifications_diff:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={
                    "name": result["old_state"].get("name"),
                    "resource_id": result["old_state"].get("resource_id"),
                    "notifications": result["old_state"].get("notifications"),
                },
                desired_state={
                    "name": name,
                    "resource_id": name + "-notifications",
                    "notifications": notifications,
                },
            )
            if resource_id:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=name
                )
            else:
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type=STATE_NAME, name=name
                )
            return result
        notification_ret = (
            await hub.exec.boto3.client.s3.put_bucket_notification_configuration(
                ctx, Bucket=name, NotificationConfiguration=notifications
            )
        )
        if not notification_ret["result"]:
            result["result"] = False
            result["comment"] = notification_ret["comment"]
            return result

        bucket_notification_result = (
            await hub.exec.boto3.client.s3.get_bucket_notification_configuration(
                ctx, Bucket=name
            )
        )
        if bucket_notification_result["result"]:
            result[
                "new_state"
            ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_notification_to_present(
                bucket_notification_result["ret"], name
            )
        else:
            result["comment"] = bucket_notification_result["comment"]
            result["result"] = False
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    This action enables you to delete multiple notification configurations from a bucket using a single HTTP request.
    If you know the object keys that you want to delete, then this action provides a suitable alternative to sending
    individual delete requests, reducing per-request overhead. For each configuration, Amazon S3 performs a delete
    action and returns the result of that delete, success, or failure, in the response.

    Args:
        hub:
        ctx:
        name(Text): Name of the bucket on which notification needs to be configured.
        resource_id(Text, optional): Name of the bucket and 'notifications' keyword with a seperator '-'.

    Returns:
        Dict[str, Any]

    Request Syntax:
        [bucket_name-notification]:
              aws.s3.bucket_notification.absent:
                - name: [string]
                - resource_id: [string]


    Examples:

        .. code-block:: sls

            test-bucket-notification:
              aws.s3.bucket_notification.absent:
                - name: test-bucket
                - resource_id: test-bucket-notifications
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if resource_id:
        name = resource_id.split("-notifications")[0]

    # check if the bucket exists
    ret = await hub.exec.boto3.client.s3.head_bucket(ctx, Bucket=name)

    if not ret["result"]:
        if BUCKET_NOT_EXISTS_ERROR_MESSAGE in ret["comment"]:
            result["comment"] = (f"aws.s3.bucket with name: {name} does not exists",)
        else:
            result["comment"] = ret["comment"]
        result["result"] = False
        return result

    # Get current notification configurations attached to the bucket
    bucket_notification_response = (
        await hub.exec.boto3.client.s3.get_bucket_notification_configuration(
            ctx, Bucket=name
        )
    )
    if bucket_notification_response["result"]:
        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_notification_to_present(
            bucket_notification_response["ret"], name
        )
    else:
        result["comment"] = bucket_notification_response["comment"]
        result["result"] = False
        return result

    if ctx.get("test", False):
        result["comment"] = (
            f"Would remove aws.s3.bucket_notification from bucket:'{name}'.",
        )
        return result
    if not result["old_state"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result
    notification_ret = (
        await hub.exec.boto3.client.s3.put_bucket_notification_configuration(
            ctx, Bucket=name, NotificationConfiguration={}
        )
    )
    if not notification_ret["result"]:
        result["result"] = False
        result["comment"] = notification_ret["comment"]
        return result

    after_bucket_notification_response = (
        await hub.exec.boto3.client.s3.get_bucket_notification_configuration(
            ctx, Bucket=name
        )
    )
    if after_bucket_notification_response["result"]:
        result[
            "new_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_notification_to_present(
            after_bucket_notification_response["ret"], name
        )
    else:
        result["comment"] = after_bucket_notification_response["comment"]
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Obtain S3 bucket notifications for each bucket under the given context for any user.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_notification
    """
    result = {}
    # To describe the notification configurations of all the buckets, we first need to list all the buckets,
    # then get the notification configurations of each bucket
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:

        bucket_name = bucket.get("Name")

        # get notification configuration for each bucket
        try:
            bucket_notification_response = (
                await hub.exec.boto3.client.s3.get_bucket_notification_configuration(
                    ctx, Bucket=bucket_name
                )
            )
            if bucket_notification_response["result"]:

                bucket_notification_response["ret"].pop("ResponseMetadata", None)

                if bucket_notification_response["ret"]:
                    translated_resource = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_notification_to_present(
                        bucket_notification_response["ret"], bucket_name
                    )

                    result[bucket_name + "-notifications"] = {
                        "aws.s3.bucket_notification.present": [
                            {parameter_key: parameter_value}
                            for parameter_key, parameter_value in translated_resource.items()
                        ]
                    }

                else:
                    hub.log.warning(
                        f"Notifications configuration not present for s3 bucket '{bucket_name}'."
                        f"Describe will skip this bucket and continue."
                    )

            else:
                hub.log.warning(
                    f"Could not get attached notification configuration for bucket '{bucket_name}' with error"
                    f" {bucket_notification_response['comment']} . Describe will skip this bucket and continue."
                )

        except Exception as e:
            hub.log.warning(
                f"Could not get attached notification configuration for bucket '{bucket_name}' with error"
                f" {e} . Describe will skip this bucket and continue."
            )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def update_security_group_rule(
    hub,
    ctx,
    before: Dict[str, Any],
    cidr_ipv4: str,
    cidr_ipv6: str,
    description: str,
    from_port: int,
    to_port: int,
    group_id: str,
    ip_protocol: str,
    prefix_list_id: str,
    referenced_group_info: Dict,
    resource_id: str,
):
    """
    Updates the Security group rule
    Args:
        before: existing resource
        ip_protocol(str): The IP protocol name (tcp , udp , icmp , icmpv6 ) or number (see Protocol Numbers ). [VPC only] Use -1 to specify all protocols.
                          When authorizing security group rules, specifying -1 or a protocol number other than tcp , udp , icmp , or icmpv6 allows traffic
                          on all ports, regardless of any port range you specify. For tcp , udp , and icmp , you must specify a port range. For icmpv6 ,
                          the port range is optional; if you omit the port range, traffic for all types and codes is allowed.
        from_port(int): The start of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 type number. A value of -1 indicates all ICMP/ICMPv6 types.
                        If you specify all ICMP/ICMPv6 types, you must specify all codes.
        to_port(int): The end of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 code. A value of -1 indicates all ICMP/ICMPv6 codes.
                      If you specify all ICMP/ICMPv6 types, you must specify all codes.
        cidr_ipv4(str, optional): The IPv4 CIDR range. You can either specify a CIDR range or a source security group, not both.
                        To specify a single IPv4 address, use the /32 prefix length.
        resource_id(Text, optional): AWS Security group rule ID.
        cidr_ipv6(str, optional): The IPv6 CIDR range. You can either specify a CIDR range or a source security group, not both.
                                  To specify a single IPv6 address, use the /128 prefix length.
        prefix_list_id(str, optional): The ID of the prefix.
        referenced_group_info(Dict, optional): The security group and Amazon Web Services account ID pairs.
        description(str, optional): The description of the security group rule.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    updated_payload = {}
    updated_rule = {}
    new_modified_rule = dict(SecurityGroupRuleId=None, SecurityGroupRule=None)
    new_modified_rule["SecurityGroupRuleId"] = resource_id
    resource_parameters = OrderedDict(
        {
            "IpProtocol": ip_protocol,
            "FromPort": from_port,
            "ToPort": to_port,
            "CidrIpv4": cidr_ipv4,
            "CidrIpv6": cidr_ipv6,
            "PrefixListId": prefix_list_id,
            "Description": description,
        }
    )
    for key, value in resource_parameters.items():
        if key in before.keys():
            updated_payload[key] = resource_parameters[key]
            if (value is not None) and value != before[key]:
                updated_rule[key] = resource_parameters[key]
        elif resource_parameters[key]:
            updated_payload[key] = resource_parameters[key]
            updated_rule[key] = resource_parameters[key]
    # You must either specify cidr_ipv4 or referenced_group_info. You cannot change source from cidr to reference
    # group id or vice versa
    if before.get(
        "CidrIpv4"
    ) is None and not hub.tool.aws.state_comparison_utils.compare_dicts(
        referenced_group_info,
        before.get("ReferencedGroupInfo"),
    ):
        updated_rule["ReferencedGroupId"] = referenced_group_info.get("GroupId")
    if updated_rule:
        if before.get("CidrIpv4") is None and referenced_group_info is not None:
            updated_payload["ReferencedGroupId"] = referenced_group_info.get("GroupId")
        new_modified_rule["SecurityGroupRule"] = updated_payload
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ec2.modify_security_group_rules(
                ctx,
                GroupId=group_id,
                SecurityGroupRules=[new_modified_rule],
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, updated_rule)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "IpProtocol": "ip_protocol",
            "FromPort": "from_port",
            "ToPort": "to_port",
            "CidrIpv4": "cidr_ipv4",
            "CidrIpv6": "cidr_ipv6",
            "PrefixListId": "prefix_list_id",
            "Description": "description",
            "ReferencedGroupId": "referenced_group_id",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.budgets.create_budget
hub.exec.boto3.client.budgets.delete_budget
hub.exec.boto3.client.budgets.describe_budget
hub.exec.boto3.client.budgets.update_budget
resource = hub.exec.boto3.client.budgets.create_budget(ctx, AnomalyMonitor)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    budget_name: str,
    time_unit: str,
    budget_type: str,
    budget_limit: make_dataclass("Spend", [("Amount", str), ("Unit", str)]) = None,
    planned_budget_limits: Dict[
        str, make_dataclass("Spend", [("Amount", str), ("Unit", str)])
    ] = None,
    cost_filters: Dict[str, List[str]] = None,
    cost_types: make_dataclass(
        "CostTypes",
        [
            ("IncludeTax", bool, field(default=None)),
            ("IncludeSubscription", bool, field(default=None)),
            ("UseBlended", bool, field(default=None)),
            ("IncludeRefund", bool, field(default=None)),
            ("IncludeCredit", bool, field(default=None)),
            ("IncludeUpfront", bool, field(default=None)),
            ("IncludeRecurring", bool, field(default=None)),
            ("IncludeOtherSubscription", bool, field(default=None)),
            ("IncludeSupport", bool, field(default=None)),
            ("IncludeDiscount", bool, field(default=None)),
            ("UseAmortized", bool, field(default=None)),
        ],
    ) = None,
    time_period: make_dataclass(
        "TimePeriod",
        [
            ("Start", datetime, field(default=None)),
            ("End", datetime, field(default=None)),
        ],
    ) = None,
    auto_adjust_data: make_dataclass(
        "AutoAdjustData",
        [
            ("AutoAdjustType", str),
            (
                "HistoricalOptions",
                make_dataclass(
                    "HistoricalOptions",
                    [
                        ("BudgetAdjustmentPeriod", int),
                        ("LookBackAvailablePeriods", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("LastAutoAdjustTime", str, field(default=None)),
        ],
    ) = None,
    notifications_with_subscribers: List[
        make_dataclass(
            "NotificationWithSubscribers",
            [
                (
                    "Notification",
                    make_dataclass(
                        "Notification",
                        [
                            ("NotificationType", str),
                            ("ComparisonOperator", str),
                            ("Threshold", float),
                            ("ThresholdType", str, field(default=None)),
                            ("NotificationState", str, field(default=None)),
                        ],
                    ),
                ),
                (
                    "Subscribers",
                    List[
                        make_dataclass(
                            "Subscriber", [("SubscriptionType", str), ("Address", str)]
                        )
                    ],
                ),
            ],
        )
    ] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a budget and, if included, notifications and subscribers.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): Resource id to identify the resource.
        budget_name(Text): The name of a budget. The name must be unique within an account. The : and \ characters aren't allowed in BudgetName.
        time_unit(Text): The length of time until a budget resets the actual and forecasted spend.
            'DAILY'|'MONTHLY'|'QUARTERLY'|'ANNUALLY'
        budget_type(Text): Specifies whether this budget tracks costs, usage, RI utilization, RI coverage, Savings Plans utilization, or Savings Plans coverage.
        budget_limit(Dict, optional): The total amount of cost, usage, RI utilization, RI coverage, Savings Plans utilization, or Savings Plans coverage
            that you want to track with your budget. BudgetLimit is required for cost or usage budgets, but optional for RI or Savings Plans utilization or coverage budgets.
            RI and Savings Plans utilization or coverage budgets default to 100 . This is the only valid value for RI or Savings Plans utilization or
            coverage budgets. You can't use BudgetLimit with PlannedBudgetLimits for CreateBudget and UpdateBudget actions.
            * Amount (Text) -- [REQUIRED] The cost or usage amount that's associated with a budget forecast, actual spend, or budget threshold.
            * Unit (Text) -- [REQUIRED] The unit of measurement that's used for the budget forecast, actual spend, or budget threshold, such as USD or GBP.
        planned_budget_limits(Dict[str, Any], optional): The total amount of cost, usage, RI utilization, RI coverage, Savings Plans utilization, or
            Savings Plans coverage that you want to track with your budget.  BudgetLimit is required for
            cost or usage budgets, but optional for RI or Savings Plans utilization or coverage budgets. RI
            and Savings Plans utilization or coverage budgets default to 100. This is the only valid value
            for RI or Savings Plans utilization or coverage budgets. You can't use BudgetLimit with
            PlannedBudgetLimits for CreateBudget and UpdateBudget actions.
            * Amount (str): The cost or usage amount that's associated with a budget forecast, actual spend, or budget threshold.
            * Unit (str): The unit of measurement that's used for the budget forecast, actual spend, or budget threshold, such as USD or GBP.
        cost_filters(Dict[str, List[str]], optional): The cost filters, such as Region , Service , member account , Tag , or Cost Category , that are applied to a budget.
        cost_types(Dict[str, Any], optional): The types of costs that are included in this COST budget.  USAGE, RI_UTILIZATION, RI_COVERAGE,
            SAVINGS_PLANS_UTILIZATION, and SAVINGS_PLANS_COVERAGE budgets do not have CostTypes.
            * IncludeTax (bool, optional): Specifies whether a budget includes taxes. The default value is true.
            * IncludeSubscription (bool, optional): Specifies whether a budget includes subscriptions. The default value is true.
            * UseBlended (bool, optional): Specifies whether a budget uses a blended rate. The default value is false.
            * IncludeRefund (bool, optional): Specifies whether a budget includes refunds. The default value is true.
            * IncludeCredit (bool, optional): Specifies whether a budget includes credits. The default value is true.
            * IncludeUpfront (bool, optional): Specifies whether a budget includes upfront RI costs. The default value is true.
            * IncludeRecurring (bool, optional): Specifies whether a budget includes recurring fees such as monthly RI fees. The default value istrue.
            * IncludeOtherSubscription (bool, optional): Specifies whether a budget includes non-RI subscription costs. The default value is true.
            * IncludeSupport (bool, optional): Specifies whether a budget includes support subscription fees. The default value is true.
            * IncludeDiscount (bool, optional): Specifies whether a budget includes discounts. The default value is true.
            * UseAmortized (bool, optional): Specifies whether a budget uses the amortized rate. The default value is false.
        time_period (Dict[str, Any], optional): The period of time that's covered by a budget. You setthe start date and end date. The start
            date must come before the end date. The end date must come before 06/15/87 00:00 UTC.  If you
            create your budget and don't specify a start date, Amazon Web Services defaults to the start of
            your chosen time period (DAILY, MONTHLY, QUARTERLY, or ANNUALLY). For example, if you created
            your budget on January 24, 2018, chose DAILY, and didn't set a start date, Amazon Web Services
            set your start date to 01/24/18 00:00 UTC. If you chose MONTHLY, Amazon Web Services set your
            start date to 01/01/18 00:00 UTC. If you didn't specify an end date, Amazon Web Services set
            your end date to 06/15/87 00:00 UTC. The defaults are the same for the Billing and Cost
            Management console and the API.  You can change either date with the UpdateBudget operation.
            After the end date, Amazon Web Services deletes the budget and all the associated notifications
            and subscribers.
            * Start (datetime, optional): The start date for a budget. If you created your budget and didn't specify a start date, Amazon
                Web Services defaults to the start of your chosen time period (DAILY, MONTHLY, QUARTERLY, or
                ANNUALLY). For example, if you created your budget on January 24, 2018, chose DAILY, and didn't
                set a start date, Amazon Web Services set your start date to 01/24/18 00:00 UTC. If you chose
                MONTHLY, Amazon Web Services set your start date to 01/01/18 00:00 UTC. The defaults are the
                same for the Billing and Cost Management console and the API. You can change your start date
                with the UpdateBudget operation.
            * End (datetime, optional): The end date for a budget. If you didn't specify an end date, Amazon Web Services set your end
                date to 06/15/87 00:00 UTC. The defaults are the same for the Billing and Cost Management
                console and the API. After the end date, Amazon Web Services deletes the budget and all the
                associated notifications and subscribers. You can change your end date with the UpdateBudget
                operation.
        auto_adjust_data (Dict[str, Any], optional):: The parameters that determine the budget amount for an auto-adjusting budget.
            * AutoAdjustType (str):-- [REQUIRED] The string that defines whether your budget auto-adjusts based on historical or forecasted data.
            * HistoricalOptions (Dict[str, Any], optional): -- The parameters that define or describe the historical data that your auto-adjusting budget is based on.
                * BudgetAdjustmentPeriod (Integer) -- [REQUIRED] The number of budget periods included in the moving-average calculation that determines your auto-adjusted budget amount.
                    The maximum value depends on the TimeUnit granularity of the budget:
                    For the DAILY granularity, the maximum value is 60 .
                    For the MONTHLY granularity, the maximum value is 12 .
                    For the QUARTERLY granularity, the maximum value is 4 .
                    For the ANNUALLY granularity, the maximum value is 1 .
                * LookBackAvailablePeriods (Integer) -- The integer that describes how many budget periods in your BudgetAdjustmentPeriod are included in the calculation
                    of your current BudgetLimit. If the first budget period in your BudgetAdjustmentPeriod has no cost data, then that budget period isn’t included in the average
                    that determines your budget limit. For example, if you set BudgetAdjustmentPeriod as 4 quarters, but your account had no cost data in the first quarter,
                    then only the last three quarters are included in the calculation. In this scenario, LookBackAvailablePeriods returns 3. You can’t set your own LookBackAvailablePeriods.
                The value is automatically calculated from the BudgetAdjustmentPeriod and your historical cost data.
        notifications_with_subscribers(List[Dict[str, Any]], optional): A notification that you want to associate with a budget. A budget can have up to five notifications,
            and each notification can have one SNS subscriber and up to 10 email subscribers. If you include notifications and subscribers in your CreateBudget call,
            Amazon Web Services creates the notifications and subscribers for you.
            A notification with subscribers. A notification can have one SNS subscriber and up to 10 email subscribers, for a total of 11 subscribers.
            * Notification (Dict[str, Any]): -- [REQUIRED] The notification that's associated with a budget.
                * NotificationType (str) -- [REQUIRED] Specifies whether the notification is for how much you have spent (ACTUAL ) or for how much that you're forecasted to spend (FORECASTED ).
                * ComparisonOperator (str) -- [REQUIRED] The comparison that's used for this notification.
                * Threshold (float) -- [REQUIRED] The threshold that's associated with a notification. Thresholds are always a percentage, and many customers find value being
                    alerted between 50% - 200% of the budgeted amount. The maximum limit for your threshold is 1,000,000% above the budgeted amount.
                * ThresholdType (str) -- The type of threshold for a notification. For ABSOLUTE_VALUE thresholds, Amazon Web Services notifies you when you go over or are
                    forecasted to go over your total cost threshold. For PERCENTAGE thresholds, Amazon Web Services notifies you when you go over or are forecasted to go over a
                    certain percentage of your forecasted spend. For example, if you have a budget for 200 dollars and you have a PERCENTAGE threshold of 80%, Amazon Web Services notifies
                    you when you go over 160 dollars.
                * NotificationState (string) -- Specifies whether this notification is in alarm. If a budget notification is in the ALARM state, you passed the set threshold for the budget.
            * Subscribers (List[Dict[str, Any]]): -- [REQUIRED] A list of subscribers who are subscribed to this notification. The subscriber to a budget notification. The subscriber consists of a
                subscription type and either an Amazon SNS topic or an email address. For example, an email subscriber has the following parameters:
                    A subscriptionType of EMAIL
                    An address of example@example.com
                * SubscriptionType (Text) -- [REQUIRED] The type of notification that Amazon Web Services sends to a subscriber.
                * Address (Text) -- [REQUIRED] The address that Amazon Web Services sends budget notifications to, either an SNS topic or an email. When you create a subscriber,
                    the value of Address can't contain line breaks.


    Request Syntax:
        [budget-resource-id]:
          aws.budgets.budget.present:
          - resource_id: 'string'
          - budget_name: 'string'
          - time_unit: 'string'
          - budget_type: 'string'
          - budget_limit: 'dict'
          - cost_filters: 'dict'
          - cost_types: 'dict'
          - time_period: 'dict'
          - auto_adjust_data: 'dict'
          - notifications_with_subscribers: 'dict'

    Returns:
        Dict[str, str]

    Examples:

        .. code-block:: sls

            cost_budget1234:
              aws.budgets.budget.present:
              - budget_name: new-test-budget
              - budget_limit:
                  Amount: '700.0'
                  Unit: USD
              - cost_filters: {}
              - cost_types:
                  IncludeCredit: false
                  IncludeDiscount: true
                  IncludeOtherSubscription: true
                  IncludeRecurring: true
                  IncludeRefund: false
                  IncludeSubscription: true
                  IncludeSupport: true
                  IncludeTax: true
                  IncludeUpfront: true
                  UseAmortized: false
                  UseBlended: false
              - time_unit: MONTHLY
              - time_period:
                  End: 2087-06-15 05:30:00+05:30
                  Start: 2022-06-01 05:30:00+05:30
              - budget_type: COST

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    resource_updated = False

    budget_parameters = {
        "BudgetName": budget_name,
        "CostFilters": cost_filters,
        "CostTypes": cost_types,
        "TimeUnit": time_unit,
        "TimePeriod": time_period,
        "BudgetType": budget_type,
    }

    if budget_limit:
        budget_parameters["BudgetLimit"] = budget_limit
    else:
        budget_parameters["PlannedBudgetLimits"] = planned_budget_limits
    if auto_adjust_data:
        budget_parameters["AutoAdjustData"] = auto_adjust_data

    if resource_id:
        before = await hub.exec.boto3.client.budgets.describe_budget(
            ctx, AccountId=account_id, BudgetName=resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.exec.aws.budgets.budget.update_budget(
            ctx,
            account_id=account_id,
            before=before["ret"]["Budget"],
            budget_limit=budget_limit,
            planned_budget_limits=planned_budget_limits,
            cost_filters=cost_filters,
            cost_types=cost_types,
            time_unit=time_unit,
            time_period=time_period,
            budget_type=budget_type,
            auto_adjust_data=auto_adjust_data,
        )
        result["comment"] += update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            plan_state = update_ret["ret"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.budgets.budget", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.budgets.budget", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "budget_name": budget_name,
                    "budget_limit": budget_limit,
                    "cost_filters": cost_filters,
                    "cost_types": cost_types,
                    "time_unit": time_unit,
                    "time_period": time_period,
                    "budget_type": budget_type,
                    "auto_adjust_data": auto_adjust_data,
                    "notifications_with_subscribers": notifications_with_subscribers,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.budgets.budget", name=name
            )
            return result
        ret = await hub.exec.boto3.client.budgets.create_budget(
            ctx,
            AccountId=account_id,
            Budget=budget_parameters,
            NotificationsWithSubscribers=notifications_with_subscribers,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.budgets.budget", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.budgets.describe_budget(
                ctx, AccountId=account_id, BudgetName=budget_name
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
                ctx,
                account_id=account_id,
                raw_resource=after["ret"]["Budget"],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    **Autogenerated function**

    Deletes a budget. You can delete your budget at any time.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, Optional):  Budget id to identify the resource. Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.budgets.budget.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget", name=name
        )
        return result

    before = await hub.exec.boto3.client.budgets.describe_budget(
        ctx, AccountId=account_id, BudgetName=name
    )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.budgets.budget", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        ret = await hub.exec.boto3.client.budgets.delete_budget(
            ctx, AccountId=account_id, BudgetName=name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.budgets.budget", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Lists the budgets that are associated with an account.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.budgets.budget
    """
    result = {}
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    ret = await hub.exec.boto3.client.budgets.describe_budgets(
        ctx, AccountId=account_id
    )

    if not ret["result"]:
        hub.log.debug(f"Could not describe budgets {ret['comment']}")
        return {}

    for budget in ret["ret"]["Budgets"]:
        resource_id = budget.get("BudgetName")
        resource_translated = (
            await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
                ctx,
                account_id=account_id,
                raw_resource=budget,
                idem_resource_name=resource_id,
            )
        )
        result[resource_id] = {
            "aws.budgets.budget.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Any
from typing import Dict
from typing import List


async def update_transit_gateway_vpc_attachment(
    hub,
    ctx,
    transit_gateway_vpc_attachment_id: str,
    old_subnets: List,
    new_subnets: List,
    old_options: Dict[str, Any],
    new_options: Dict[str, Any],
):
    """
    Update options and subnets of a transit_gateway vpc attachment. This function compares the existing(old) options
    or subnets and the new options or subnets. subnets that are in the new new_subnets but not in the
    old_subnets will be associated to transit gateway vpc attachment. subnets that are in the
    old_subnets but not in the new_subnets will be disassociated from transit gateway vpc attachment.

    Args:
        hub:
        ctx:
        transit_gateway_vpc_attachment_id(Text): The AWS resource id of the existing transit gateway vpc attachment
        old_subnets(List):  subnets of existing transit gateway vpc attachment
        new_subnets(List): new subnets to update
        old_options(dict): options of existing transit gateway vpc attachment
        new_options(dict): options to be updated

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    if set(old_subnets) == set(new_subnets) and (
        new_options is None
        or hub.tool.aws.state_comparison_utils.compare_dicts(new_options, old_options)
    ):
        return result

    subnets_to_add = list(set(new_subnets).difference(old_subnets))
    subnets_to_remove = list(set(old_subnets).difference(new_subnets))

    if ctx.get("test", False):
        result["comment"] = result["comment"] + (
            f"Would update transit gateway vpc attachment {transit_gateway_vpc_attachment_id}",
        )
    else:
        ret = await hub.exec.boto3.client.ec2.modify_transit_gateway_vpc_attachment(
            ctx,
            TransitGatewayAttachmentId=transit_gateway_vpc_attachment_id,
            AddSubnetIds=subnets_to_add,
            RemoveSubnetIds=subnets_to_remove,
            Options=new_options,
        )
        result["result"] = ret["result"]
        if not ret["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
    result["ret"] = {
        "subnet_ids": new_subnets,
        "options": new_options,
    }
    return result


async def get_transit_gateway_vpc_attachment_by_id(
    hub, ctx, transit_gateway_vpc_attachment_id
):
    ret = {}
    result = dict(comment=(), result=True, ret=None)
    before = await hub.exec.boto3.client.ec2.describe_transit_gateway_vpc_attachments(
        ctx, TransitGatewayAttachmentIds=[transit_gateway_vpc_attachment_id]
    )
    result["result"] = before["result"]
    result["comment"] = before["comment"]
    if before["result"]:
        for transit_gateway_vpc_attachment in before["ret"][
            "TransitGatewayVpcAttachments"
        ]:
            ret = transit_gateway_vpc_attachment
    result["ret"] = ret
    return result



async def lookup(hub, ctx, name: str, service: str, resource: str, filter: str = "[]"):
    """
    A state to look up a resource that is not managed.
    It's data will be available to other states via the arg_bind requisite interface.

    :param hub:
    :param ctx:
    :param name: The name of the state
    :param service: The name of the service under the boto3 client to use
    :param resource: The name of the collection under the service to search
    :param filter: A JMESPath

    .. code-block:: sls

        my_unmanaged_resource:
            aws.data.lookup:
                service: ec2
                resource: vpc
                filter: *[]
    """
    ret = dict(comment="", old_state={}, new_state={}, name=name, result=True)
    resource = await hub.exec.aws.data.lookup(
        ctx, service=service, resource=resource, filter=filter
    )
    ret["comment"] = resource.comment
    ret["result"] = resource.result
    ret["old_state"] = resource.ret
    ret["new_state"] = resource.ret
    return ret



"""
hub.exec.boto3.client.config.put_config_rule
hub.exec.boto3.client.config.delete_config_rule
hub.exec.boto3.client.config.describe_config_rules

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    scope: make_dataclass(
        "Scope",
        [
            ("ComplianceResourceTypes", List[str], field(default=None)),
            ("TagKey", str, field(default=None)),
            ("TagValue", str, field(default=None)),
            ("ComplianceResourceId", str, field(default=None)),
        ],
    ) = None,
    source: make_dataclass(
        "Source",
        [
            ("Owner", str, field(default=None)),
            ("SourceIdentifier", str, field(default=None)),
            (
                "SourceDetails",
                List[
                    make_dataclass(
                        "SourceDetail",
                        [
                            ("EventSource", str, field(default=None)),
                            ("MessageType", str, field(default=None)),
                            ("MaximumExecutionFrequency", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            (
                "CustomPolicyDetails",
                make_dataclass(
                    "CustomPolicyDetail",
                    [
                        ("PolicyRuntime", str, field(default=None)),
                        ("PolicyText", str, field(default=None)),
                        ("EnableDebugLogDelivery", bool, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    max_execution_frequency: str = None,
    input_parameters: str = None,
) -> Dict[str, Any]:
    """
    Adds or updates Config rule for evaluating whether your Amazon Web Services resources comply with your desired
    configurations. For more information about rules, please see AWS Config services
    Args:
        name(Text): An Idem name of the rule.
        resource_id(Text, Optional): AWS Config Rule Name.
        scope(Dict[str, Any], optional): Defines which resources can trigger an evaluation for the rule. The scope can include one or
            more resource types, a combination of one resource type and one resource ID, or a combination of
            a tag key and value. Specify a scope to constrain the resources that can trigger an evaluation
            for the rule. If you do not specify a scope, evaluations are triggered when any resource in the
            recording group changes.  The scope can be empty.
                * ComplianceResourceTypes (List[str], optional): The resource types of only those Amazon Web Services resources that you want to trigger an
                    evaluation for the rule. You can only specify one type if you also specify a resource ID for
                    ComplianceResourceId.
                * TagKey (str, optional): The tag key that is applied to only those Amazon Web Services resources that you want to trigger
                    an evaluation for the rule.
                * TagValue (str, optional): The tag value applied to only those Amazon Web Services resources that you want to trigger an
                    evaluation for the rule. If you specify a value for TagValue, you must also specify a value for
                    TagKey.
                * ComplianceResourceId (str, optional): The ID of the only Amazon Web Services resource that you want to trigger an evaluation for the
                    rule. If you specify a resource ID, you must specify one resource type for
                    ComplianceResourceTypes.
        source (Dict[str, Any]): Provides the rule owner (Amazon Web Services or customer), the rule identifier, and the
            notifications that cause the function to evaluate your Amazon Web Services resources.
            * Owner (str): Indicates whether Amazon Web Services or the customer owns and manages the Config rule. Config
                Managed Rules are predefined rules owned by Amazon Web Services. For more information, see
                Config Managed Rules in the Config developer guide. Config Custom Rules are rules that you can
                develop either with Guard (CUSTOM_POLICY) or Lambda (CUSTOM_LAMBDA). For more information, see
                Config Custom Rules  in the Config developer guide.
            * SourceIdentifier (str, optional): For Config Managed rules, a predefined identifier from a list. For example, IAM_PASSWORD_POLICY
                is a managed rule. To reference a managed rule, see List of Config Managed Rules. For Config
                Custom Lambda rules, the identifier is the Amazon Resource Name (ARN) of the rule's Lambda
                function, such as arn:aws:lambda:us-east-2:123456789012:function:custom_rule_name. For Config
                Custom Policy rules, this field will be ignored.
            * SourceDetails (List[Dict[str, Any]], optional): Provides the source and the message types that cause Config to evaluate your Amazon Web Services
                resources against a rule. It also provides the frequency with which you want Config to run
                evaluations for the rule if the trigger type is periodic. If the owner is set to CUSTOM_POLICY,
                the only acceptable values for the Config rule trigger message type are
                ConfigurationItemChangeNotification and OversizedConfigurationItemChangeNotification.
                * EventSource (str, optional): The source of the event, such as an Amazon Web Services service, that triggers Config to
                    evaluate your Amazon Web Services resources.
                * MessageType (str, optional): The type of notification that triggers Config to run an evaluation for a rule. You can specify
                    the following notification types:
                        ConfigurationItemChangeNotification - Triggers an evaluation when Config delivers a configuration item as a result of a resource change.
                        OversizedConfigurationItemChangeNotification - Triggers an evaluation when Config delivers an
                            oversized configuration item. Config may generate this notification type when a resource changes
                            and the notification exceeds the maximum size allowed by Amazon SNS.
                        ScheduledNotification - Triggers a periodic evaluation at the frequency specified for MaximumExecutionFrequency.
                        ConfigurationSnapshotDeliveryCompleted - Triggers a periodic evaluation when Config delivers a
                            configuration snapshot.
                    If you want your custom rule to be triggered by configuration changes,
                    specify two SourceDetail objects, one for ConfigurationItemChangeNotification and one for
                    OversizedConfigurationItemChangeNotification.
                * MaximumExecutionFrequency (str, optional): The frequency at which you want Config to run evaluations for a custom rule with a periodic
                    trigger. If you specify a value for MaximumExecutionFrequency, then MessageType must use the
                    ScheduledNotification value.  By default, rules with a periodic trigger are evaluated every 24
                    hours. To change the frequency, specify a valid value for the MaximumExecutionFrequency
                    parameter. Based on the valid value you choose, Config runs evaluations once for each valid
                    value. For example, if you choose Three_Hours, Config runs evaluations once every three hours.
                    In this case, Three_Hours is the frequency of this rule.
            * CustomPolicyDetails (Dict[str, Any], optional): Provides the runtime system, policy definition, and whether debug logging is enabled. Required
                when owner is set to CUSTOM_POLICY.
                * PolicyRuntime (str): The runtime system for your Config Custom Policy rule. Guard is a policy-as-code language that
                    allows you to write policies that are enforced by Config Custom Policy rules. For more
                    information about Guard, see the Guard GitHub Repository.
                * PolicyText (str): The policy definition containing the logic for your Config Custom Policy rule.
                * EnableDebugLogDelivery (bool, optional): The boolean expression for enabling debug logging for your Config Custom Policy rule. The
                    default value is false.
        max_execution_frequency(Text, Optional): The maximum frequency with which Config runs evaluations for a rule. Default
         is 24 hours
        input_parameters(Text, Optional): A string, in JSON format, that is passed to the Config rule.
        tags (Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the Config rule.
            * (Key, optional): One part of a key-value pair that make up a tag. A key is a general label that acts like a
                category for more specific tag values.
            * (Value, optional): The optional part of a key-value pair that make up a tag. A value acts as a descriptor within a
                tag category (key).

        tags(List or Dict, Optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}. The tags for the resource. The metadata that you apply to a resource to help you categorize and
         organize them. Defaults to None

    Request syntax:
        [aws-config-rule]:
          aws.config.rule.present:
          - name: 'string'
          - resource_id: 'string'
          - scope: dict
            ComplianceResourceTypes: list
          - source: dict
            Owner: 'string'
            SourceIdentifier: 'string'

    Returns:
         None

    Examples:
        .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.rule.present:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip
              - tags:
                - Key: ENV
                  Value: Test
                - Key: Service
                  Value: TestService
              - config_rule_name: ec2-instance-no-public-ip
              - scope:
                  ComplianceResourceTypes:
                  - AWS::EC2::Instance
                  - AWS::EC2::Host
              - source:
                  Owner: AWS
                  SourceIdentifier: EC2_INSTANCE_NO_PUBLIC_IP
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    update_scope = None
    if resource_id:
        try:
            before = await hub.exec.boto3.client.config.describe_config_rules(
                ctx, ConfigRuleNames=[resource_id]
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                ctx,
                raw_resource=before["ret"]["ConfigRules"][0],
                idem_resource_name=name,
            )
            plan_state = copy.deepcopy(result["old_state"])
            # updating scope
            if scope is not None:
                update_scope = (
                    scope if scope is not None else result["old_state"]["scope"]
                )

            update_ret = await hub.exec.aws.config.rule.update_rule(
                ctx,
                resource_id=resource_id,
                before=before["ret"]["ConfigRules"][0],
                source=source,
                scope=update_scope,
                frequency=max_execution_frequency,
                input_parameters=input_parameters,
            )

            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                for key in ["max_execution_frequency", "scope", "input_parameters"]:
                    if key in update_ret["ret"]:
                        plan_state[key] = update_ret["ret"].get(key)

            if (tags is not None) and tags != result["old_state"].get("tags"):
                update_tag_ret = await hub.exec.aws.config.tag.update_tags(
                    ctx,
                    result["old_state"]["config_rule_arn"],
                    result["old_state"]["tags"],
                    tags,
                )
                result["result"] = result["result"] and update_tag_ret["result"]
                result["comment"] = result["comment"] + update_tag_ret["comment"]
                resource_updated = resource_updated or bool(update_tag_ret["result"])
                if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                    plan_state["tags"] = update_tag_ret["ret"]
            if not resource_updated:
                result["comment"] = result["comment"] + (f"{name} already exists",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        try:
            if ctx.get("test", False):
                desired_state_payload = {
                    "name": name,
                    "config_rule_name": name,
                    "source": source,
                    "tags": tags,
                }
                if scope:
                    desired_state_payload["scope"] = scope
                if input_parameters:
                    desired_state_payload["input_parameters"] = input_parameters
                if max_execution_frequency:
                    desired_state_payload[
                        "max_execution_frequency"
                    ] = max_execution_frequency
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state_payload
                )
                result["comment"] = (f"Would create aws.config.rule {name}",)
                return result
            config_rule_payload = {"ConfigRuleName": name, "Source": source}
            if scope:
                config_rule_payload["Scope"] = scope
            if max_execution_frequency:
                config_rule_payload[
                    "MaximumExecutionFrequency"
                ] = max_execution_frequency
            if input_parameters:
                config_rule_payload["InputParameters"] = input_parameters
            ret = await hub.exec.boto3.client.config.put_config_rule(
                ctx,
                ConfigRule=config_rule_payload,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"] + ret["comment"]
                return result
            resource_id = name
            result["comment"] = result["comment"] + (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}")
            result["result"] = False
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.config.describe_config_rules(
                ctx, ConfigRuleNames=[resource_id]
            )
            result[
                "new_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                ctx,
                raw_resource=after["ret"]["ConfigRules"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Deletes the specified Config rule and all of its evaluation results. Config sets the state of a rule to DELETING
    until the deletion is complete. You cannot update a rule while it is in this state. If you make a PutConfigRule or
    DeleteConfigRule request for the rule, you will receive a ResourceInUseException.

    Args:
        name(Text): An Idem name of the rule.
        resource_id(Text, optional): AWS Config Rule Name. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.rule.absent:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.rule", name=name
        )
        return result
    before = await hub.exec.boto3.client.config.describe_config_rules(
        ctx, ConfigRuleNames=[resource_id]
    )
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.rule", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
            ctx, raw_resource=before["ret"]["ConfigRules"][0], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.rule", name=name
        )
        return result
    else:
        try:
            if before["ret"]["ConfigRules"][0]["ConfigRuleState"] == "DELETING":
                result["comment"] = (
                    f"aws.config.rule '{name}' is still in deleting state.",
                )
            else:
                result[
                    "old_state"
                ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                    ctx,
                    raw_resource=before["ret"]["ConfigRules"][0],
                    idem_resource_name=name,
                )

                ret = await hub.exec.boto3.client.config.delete_config_rule(
                    ctx, ConfigRuleName=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.config.rule", name=name
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Return details about your Config rules.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.config.rule
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_config_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe Rules {ret['comment']}")
        return {}

    for resource in ret["ret"]["ConfigRules"]:
        resource_id = resource.get("ConfigRuleName")
        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.config.rule.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def convert_raw_cloudtrail_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("Name")
    resource_parameters = OrderedDict(
        {
            "S3BucketName": "s3_bucket_name",
            "S3KeyPrefix": "s3_key_prefix",
            "SnsTopicName": "sns_topic_name",
            "IncludeGlobalServiceEvents": "include_global_service_events",
            "IsMultiRegionTrail": "is_multi_region_trail",
            "LogFileValidationEnabled": "enable_logfile_validation",
            "CloudWatchLogsLogGroupArn": "cloud_watch_logs_loggroup_arn",
            "CloudWatchLogsRoleArn": "cloud_watch_logs_role_arn",
            "KmsKeyId": "kms_key_id",
            "IsOrganizationTrail": "is_organization_trail",
            "TrailARN": "trail_arn",
            "Tags": "tags",
            "IsLogging": "is_logging",
            "EventSelectors": "event_selectors",
            "AdvancedEventSelectors": "advanced_event_selectors",
            "InsightSelectors": "insight_selectors",
        }
    )

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    ret_tag = await hub.exec.boto3.client.cloudtrail.list_tags(
        ctx, ResourceIdList=[raw_resource.get("TrailARN")]
    )
    if ret_tag["result"]:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tag.get("ret").get("ResourceTagList")[0].get("TagsList")
        )

    ret_trail_status = await hub.exec.boto3.client.cloudtrail.get_trail_status(
        ctx, Name=raw_resource.get("TrailARN")
    )

    if ret_trail_status["result"]:
        resource_translated["is_logging"] = ret_trail_status.get("ret").get("IsLogging")

    ret_event_selectors = await hub.exec.boto3.client.cloudtrail.get_event_selectors(
        ctx, TrailName=raw_resource.get("TrailARN")
    )

    if ret_event_selectors["result"]:
        resource_translated["event_selectors"] = ret_event_selectors.get("ret").get(
            "EventSelectors"
        )
        resource_translated["advanced_event_selectors"] = ret_event_selectors.get(
            "ret"
        ).get("AdvancedEventSelectors")

    ret_insight_selectors = (
        await hub.exec.boto3.client.cloudtrail.get_insight_selectors(
            ctx, TrailName=raw_resource.get("TrailARN")
        )
    )

    if ret_insight_selectors["result"]:
        resource_translated["insight_selectors"] = ret_insight_selectors.get("ret").get(
            "InsightSelectors"
        )

    return resource_translated


def update_plan_state(
    hub, plan_state: Dict[str, Any], update_ret: Dict[str, Any]
) -> Dict[str, Any]:
    if update_ret:
        for key, value in update_ret["ret"].items():
            plan_state[key] = value
    return plan_state



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.eks.create_nodegroup
hub.exec.boto3.client.eks.delete_nodegroup
hub.exec.boto3.client.eks.describe_nodegroup
hub.exec.boto3.client.eks.list_nodegroups
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.eks.cluster.absent",
        ],
    },
    "present": {
        "require": [
            "aws.eks.cluster.present",
        ],
    },
}
default_attempts = 120


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    subnets: List[str],
    node_role: str,
    resource_id: str = None,
    scaling_config: make_dataclass(
        "NodegroupScalingConfig",
        [
            ("minSize", int, field(default=None)),
            ("maxSize", int, field(default=None)),
            ("desiredSize", int, field(default=None)),
        ],
    ) = None,
    disk_size: int = None,
    instance_types: List[str] = None,
    ami_type: str = None,
    remote_access: make_dataclass(
        "RemoteAccessConfig",
        [
            ("ec2SshKey", str, field(default=None)),
            ("sourceSecurityGroups", List[str], field(default=None)),
        ],
    ) = None,
    labels: Dict[str, str] = None,
    taints: List[
        make_dataclass(
            "Taint",
            [
                ("key", str, field(default=None)),
                ("value", str, field(default=None)),
                ("effect", str, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, str] = None,
    client_request_token: str = None,
    launch_template: make_dataclass(
        "LaunchTemplateSpecification",
        [
            ("name", str, field(default=None)),
            ("version", str, field(default=None)),
            ("id", str, field(default=None)),
        ],
    ) = None,
    update_config: make_dataclass(
        "NodegroupUpdateConfig",
        [
            ("maxUnavailable", int, field(default=None)),
            ("maxUnavailablePercentage", int, field(default=None)),
        ],
    ) = None,
    capacity_type: str = None,
    version: str = None,
    release_version: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a managed node group for an Amazon EKS cluster. You can only create a node group for your cluster that
    is equal to the current Kubernetes version for the cluster. All node groups are created with the latest AMI
    release version for the respective minor Kubernetes version of the cluster, unless you deploy a custom AMI using
    a launch template. For more information about using launch templates, see Launch template support. An Amazon EKS
    managed node group is an Amazon EC2 Auto Scaling group and associated Amazon EC2 instances that are managed by
    Amazon Web Services for an Amazon EKS cluster. Each node group uses a version of the Amazon EKS optimized Amazon
    Linux 2 AMI. For more information, see Managed Node Groups in the Amazon EKS User Guide.

    Args:
        name(Text): An Idem name of EKS Node group.
        cluster_name(Text): The name of the cluster to create the node group in.
        subnets(List[str]): The subnets to use for the Auto Scaling group that is created for your node group. If you
            specify launchTemplate, then don't specify  SubnetId  in your launch template, or the node group
            deployment will fail. For more information about using launch templates with Amazon EKS, see
            Launch template support in the Amazon EKS User Guide.
        node_role(Text): The Amazon Resource Name (ARN) of the IAM role to associate with your node group.
            The Amazon EKS worker node kubelet daemon makes calls to Amazon Web Services APIs on your behalf. Nodes
            receive permissions for these API calls through an IAM instance profile and associated policies. Before you
            can launch nodes and register them into a cluster, you must create an IAM role for those nodes to use when
            they are launched.
        resource_id(Text, optional): AWS EKS Node group name
        scaling_config(Dict[str, Any], optional): The scaling configuration details for the Auto Scaling group that is created for your node
            group. Defaults to None.
            * minSize (int, optional): The minimum number of nodes that the managed node group can scale in to.
            * maxSize (int, optional): The maximum number of nodes that the managed node group can scale out to. For information about
                the maximum number that you can specify, see Amazon EKS service quotas in the Amazon EKS User
                Guide.
            * desiredSize (int, optional): The current number of nodes that the managed node group should maintain.  If you use Cluster
                Autoscaler, you shouldn't change the desiredSize value directly, as this can cause the Cluster
                Autoscaler to suddenly scale up or scale down.  Whenever this parameter changes, the number of
                worker nodes in the node group is updated to the specified size. If this parameter is given a
                value that is smaller than the current number of running worker nodes, the necessary number of
                worker nodes are terminated to match the given value. When using CloudFormation, no action
                occurs if you remove this parameter from your CFN template. This parameter can be different from
                minSize in some cases, such as when starting with extra hosts for testing. This parameter can
                also be different when you want to start with an estimated number of needed hosts, but let
                Cluster Autoscaler reduce the number if there are too many. When Cluster Autoscaler is used, the
                desiredSize parameter is altered by Cluster Autoscaler (but can be out-of-date for short periods
                of time). Cluster Autoscaler doesn't scale a managed node group lower than minSize or higher
                than maxSize.
        capacity_type(Text, optional): The capacity type for your node group
        disk_size(integer, optional): The root device disk size (in GiB) for your node group instances. The default
            disk size is 20 GiB.
        instance_types(List[str], optional): Specify the instance types for a node group. If you specify a GPU instance type, be sure to
            specify AL2_x86_64_GPU with the amiType parameter. If you specify launchTemplate, then you can
            specify zero or one instance type in your launch template or you can specify 0-20 instance types
            for instanceTypes. If however, you specify an instance type in your launch template and specify
            any instanceTypes, the node group deployment will fail. If you don't specify an instance type in
            a launch template or for instanceTypes, then t3.medium is used, by default. If you specify Spot
            for capacityType, then we recommend specifying multiple values for instanceTypes. For more
            information, see Managed node group capacity types and Launch template support in the Amazon EKS
            User Guide. Defaults to None.
        ami_type(Text, optional): The AMI type for your node group. GPU instance types should use the AL2_x86_64_GPU
            AMI type. Non-GPU instances should use the AL2_x86_64 AMI type. Arm instances should use the AL2_ARM_64 AMI
            type.
        remote_access(Dict[str, Any], optional): The remote access (SSH) configuration to use with your node group. If you specify
            launchTemplate, then don't specify remoteAccess, or the node group deployment will fail. For
            more information about using launch templates with Amazon EKS, see Launch template support in
            the Amazon EKS User Guide. Defaults to None.
            * ec2SshKey (str, optional): The Amazon EC2 SSH key that provides access for SSH communication with the nodes in the managed
                node group. For more information, see Amazon EC2 key pairs and Linux instances in the Amazon
                Elastic Compute Cloud User Guide for Linux Instances.
            * sourceSecurityGroups (List[str], optional): The security groups that are allowed SSH access (port 22) to the nodes. If you specify an Amazon
                EC2 SSH key but do not specify a source security group when you create a managed node group,
                then port 22 on the nodes is opened to the internet (0.0.0.0/0). For more information, see
                Security Groups for Your VPC in the Amazon Virtual Private Cloud User Guide.
        labels(Dict[str, str], optional): The Kubernetes labels to be applied to the nodes in the node group when they are created. Defaults to None.
        taints(List[Dict[str, Any]], optional): The Kubernetes taints to be applied to the nodes in the node group. For more information, see
            Node taints on managed node groups. Defaults to None.
            * key (str, optional): The key of the taint.
            * value (str, optional): The value of the taint.
            * effect (str, optional): The effect of the taint.
        tags(Dict, optional): The metadata to apply to the node group to assist with categorization and organization.
            Each tag consists of a key and an optional value. You define both. Node group tags do not propagate to any
            other resources associated with the node group, such as the Amazon EC2 instances or subnets.
        client_request_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the
            idempotency of the request.
        launch_template(Dict[str, Any], optional): An object representing a node group's launch template specification. If specified, then do not
            specify instanceTypes, diskSize, or remoteAccess and make sure that the launch template meets
            the requirements in launchTemplateSpecification. Defaults to None.
            * name (str, optional): The name of the launch template.
            * version (str, optional): The version of the launch template to use. If no version is specified, then the template's
                default version is used.
            * id (str, optional): The ID of the launch template.
        update_config(Dict[str, Any], optional): The node group update configuration. Defaults to None.
            * maxUnavailable (int, optional): The maximum number of nodes unavailable at once during a version update. Nodes will be updated
                in parallel. This value or maxUnavailablePercentage is required to have a value.The maximum
                number is 100.
            * maxUnavailablePercentage (int, optional): The maximum percentage of nodes unavailable during a version update. This percentage of nodes
                will be updated in parallel, up to 100 nodes at once. This value or maxUnavailable is required
                to have a value.
        version(str, optional): The Kubernetes version to use for your managed nodes. By default, the Kubernetes version of the
            cluster is used, and this is the only accepted specified value. If you specify launchTemplate,
            and your launch template uses a custom AMI, then don't specify version, or the node group
            deployment will fail. For more information about using launch templates with Amazon EKS, see
            Launch template support in the Amazon EKS User Guide. Defaults to None.
        release_version(str, optional): The AMI version of the Amazon EKS optimized AMI to use with your node group. By default, the
            latest available AMI version for the node group's current Kubernetes version is used. For more
            information, see Amazon EKS optimized Amazon Linux 2 AMI versions in the Amazon EKS User Guide.
            If you specify launchTemplate, and your launch template uses a custom AMI, then don't specify
            releaseVersion, or the node group deployment will fail. For more information about using launch
            templates with Amazon EKS, see Launch template support in the Amazon EKS User Guide. Defaults to None.
        timeout(Dict, optional): Timeout configuration for creating or updating nodegroup.
            * create (Dict) -- Timeout configuration for creating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:

        [eks-nodegroup-name]:
              aws.eks.nodegroup.present:
              - cluster_name: 'string'
              - node_group_arn: 'string'
              - version: 'string'
              - release_version: 'string'
              - resource_id: 'string'
              - status: 'string'
              - capacity_type: 'ON_DEMAND' | 'SPOT',
              - instance_types:
                - 'string'
              - subnets:
                - 'string'
              - ami_type: 'string'
              - node_role: 'string'
              - disk_size: int
              - scaling_config:
                  desiredSize: int
                  maxSize: int
                  minSize: int
              - update_config:
                  maxUnavailable: int
              - tags:
                  - 'string': 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nodes1:
              aws.eks.nodegroup.present:
              - cluster_name: idem-test-cluster
              - node_group_arn: arn:aws:eks:us-west-2:000000000000:nodegroup/prod/idem-test-instance-e1a573b9-4e74-48f0-85c4-c214a8ec8ada/id123
              - version: '1.21'
              - release_version: 1.21.5-20220123
              - resource_id: 'idem-test-cluster-node-group'
              - status: ACTIVE
              - capacity_type: ON_DEMAND
              - instance_types:
                - t1.micro
              - subnets:
                - subnet-31813031
              - ami_type: AL2_x86_64
              - node_role: arn:aws:iam::000000000000:role/idem-test-role-2075a427-24c2-4021-abc3-9b542834addb
              - disk_size: 20
              - scaling_config:
                  desiredSize: 2
                  maxSize: 2
                  minSize: 2
              - update_config:
                  maxUnavailable: 1
              - tags:
                  - Name: idem-test-instance-e1a573b9-4e74-48f0-85c4-c214a8ec8ada



    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.eks.describe_nodegroup(
            ctx, clusterName=cluster_name, nodegroupName=resource_id
        )
        if not before["result"]:
            if not ("ResourceNotFoundException" in before["comment"][0]):
                result["result"] = False
                result["comment"] = before["comment"]
                return result

    if before and before["ret"]:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            client_request_token = result["old_state"].get("client_request_token", None)

            # update nodegroup
            update_ret = await hub.exec.aws.eks.nodegroup.update_nodegroup(
                ctx=ctx,
                name=name,
                before=result["old_state"],
                version=version,
                release_version=release_version,
                launch_template=launch_template,
                labels=labels,
                taints=taints,
                scaling_config=scaling_config,
                update_config=update_config,
                client_request_token=client_request_token,
                timeout=timeout,
            )
            result["comment"] = update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = bool(update_ret["ret"])

            if update_ret["ret"] and ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.eks.nodegroup '{name}'",
                )
                for nodegroup_param in update_ret["ret"]:
                    plan_state[nodegroup_param] = update_ret["ret"][nodegroup_param]

            # update tags
            update_tags_ret = await hub.exec.aws.eks.tag.update_eks_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("nodegroup_arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if ctx.get("test", False) and update_tags_ret["ret"]:
                plan_state["tags"] = update_tags_ret["ret"].get("tags")
        except Exception as e1:
            result["comment"] = result["comment"] + (f"{e1.__class__.__name__}: {e1}",)
            result["result"] = False
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cluster_name": cluster_name,
                    "subnets": subnets,
                    "node_role": node_role,
                    "scaling_config": scaling_config,
                    "disk_size": disk_size,
                    "instance_types": instance_types,
                    "ami_type": ami_type,
                    "remote_access": remote_access,
                    "labels": labels,
                    "taints": taints,
                    "tags": tags,
                    "client_request_token": client_request_token,
                    "launch_template": launch_template,
                    "update_config": update_config,
                    "capacity_type": capacity_type,
                    "version": version,
                    "release_version": release_version,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.eks.create_nodegroup(
                ctx,
                clusterName=cluster_name,
                nodegroupName=name,
                scalingConfig=scaling_config,
                diskSize=disk_size,
                subnets=subnets,
                instanceTypes=instance_types,
                amiType=ami_type,
                remoteAccess=remote_access,
                nodeRole=node_role,
                labels=labels,
                taints=taints,
                tags=tags,
                clientRequestToken=client_request_token,
                launchTemplate=launch_template,
                updateConfig=update_config,
                capacityType=capacity_type,
                version=version,
                releaseVersion=release_version,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            resource_id = ret["ret"]["nodegroup"]["nodegroupName"]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=60,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.eks.nodegroup '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "nodegroup_active",
                    nodegroupName=resource_id,
                    clusterName=cluster_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.eks.describe_nodegroup(
                ctx, clusterName=cluster_name, nodegroupName=resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=after["ret"]["nodegroup"], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes an Amazon EKS node group for a cluster.

    Args:
        name(Text): An Idem name of EKS Node group.
        cluster_name(Text): The name of the cluster to create the node group in.
        resource_id(Text, optional): AWS EKS Node group name. Idem automatically considers this resource being absent
         if this field is not specified.
        timeout(Dict, optional): Timeout configuration for creating or updating cluster.
            * delete (Dict) -- Timeout configuration for deleting cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
    Request Syntax:
        [nodegroup-name]:
              aws.eks.nodegroup.absent:
                - cluster_name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nodes1:
              aws.eks.nodegroup.absent:
                - cluster_name: idem-test-cluster
                - resource_id: idem-test-nodegroup
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
        return result
    before = await hub.exec.boto3.client.eks.describe_nodegroup(
        ctx, clusterName=cluster_name, nodegroupName=resource_id
    )
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
            raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
            raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
        )
        try:
            ret = await hub.exec.boto3.client.eks.delete_nodegroup(
                ctx, clusterName=cluster_name, nodegroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            else:
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=60,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "nodegroup_deleted",
                        nodegroupName=resource_id,
                        clusterName=cluster_name,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
        except Exception as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the Amazon EKS managed node groups associated with the specified cluster in your Amazon Web Services
    account in the specified Region. Self-managed node groups are not listed.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.eks.nodegroup
    """

    result = {}
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not describe cluster {cluster_ret['comment']}")
        return {}
    cluster_node_groups = []
    try:
        for name in cluster_ret["ret"]["clusters"]:
            node_group_ret = await hub.exec.boto3.client.eks.list_nodegroups(
                ctx, clusterName=name
            )
            if not node_group_ret["result"]:
                result["comment"] = node_group_ret["comment"]
                return result
            for group in node_group_ret["ret"]["nodegroups"]:
                cluster_node_groups.append({"name": name, "group": group})
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
        return result

    for node_groups in cluster_node_groups:
        cluster_name = node_groups["name"]
        node_group = node_groups["group"]
        describe_ret = await hub.exec.boto3.client.eks.describe_nodegroup(
            ctx, clusterName=cluster_name, nodegroupName=node_group
        )
        nodegroup = describe_ret["ret"]["nodegroup"]
        resource_id = f"{nodegroup['clusterName']}-{nodegroup['nodegroupName']}"
        resource_translated = (
            hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=nodegroup, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.eks.nodegroup.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Any
from typing import Dict
from typing import List
from typing import Set

ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS = {
    "DestinationCidrBlock",
    "DestinationIpv6CidrBlock",
    "DestinationPrefixListId",
}

ALLOWED_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "EgressOnlyInternetGatewayId",
    "GatewayId",
    "InstanceId",
    "NatGatewayId",
    "TransitGatewayId",
    "LocalGatewayId",
    "CarrierGatewayId",
    "NetworkInterfaceId",
    "RouteTableId",
    "VpcPeeringConnectionId",
    "CoreNetworkArn",
}

ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "GatewayId",
    "InstanceId",
    "NetworkInterfaceId",
}


async def search_raw(hub, ctx, resource_id: str = None, filters: List = None) -> Dict:
    """
    Fetch one or more route table from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        resource_id(Text, optional): AWS route table ID to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    route_table_args = {}
    if boto3_filter:
        route_table_args["Filters"] = boto3_filter

    # Including None as RouteTableIds, throws ParamValidationError for describe_route_tables
    if resource_id:
        route_table_args["RouteTableIds"] = [resource_id]

    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx, **route_table_args)
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


def get_route_table_routes_modifications(
    hub, old_routes: List[Dict], new_routes: List[Dict]
):
    """
    Given old routes i.e, the routes of current route table and new_routes
    i.e, the routes we want to modify to, this function will compare differences between
    old and new routes and return routes to be added and removed

    Args:
        hub:
        old_routes(List):  routes of existing route table
        new_routes(List): new routes to update

    Returns:
        Dict[str, List]

    """
    result = dict(
        comment=(),
        routes_to_add=[],
        routes_to_delete=[],
        routes_to_replace=[],
        result=True,
    )
    routes_to_add = []
    routes_to_delete = []
    routes_to_replace = []
    old_routes_map = {}
    default_destination_cidr_block: str = ""
    # Creates a map of destination cidr block -> route to easily retrieve further
    for route in old_routes:
        if route.get("State") == "active":
            # Route table route allows three types of destinations
            # 1) DestinationCidrBlock,
            # 2) DestinationIpv6CidrBlock,
            # 3) DestinationPrefixListId,
            # We need to determine which type of  destination is used
            destination_cidr_key = get_common_key_in_dict(
                route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
            )
            old_routes_map[route.get(destination_cidr_key)] = route
            if route.get("Origin") == "CreateRouteTable":
                default_destination_cidr_block = route.get(destination_cidr_key)
    new_destination_cidr_blocks = list()
    for new_route in new_routes:
        # Loop through new routes and determine which destination is used
        destination_cidr_key = get_common_key_in_dict(
            new_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = new_route.get(destination_cidr_key)

        # checking for duplicate destination cidr blocks. There can be only unique destination cidr block
        if destination_cidr_block in new_destination_cidr_blocks:
            result["result"] = False
            result["comment"] = (
                "Duplicate destination cidr blocks are not allowed, they should be unique.",
            )
            return result
        else:
            new_destination_cidr_blocks.append(destination_cidr_block)
        allowed_targets = (
            ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS
            if destination_cidr_block == default_destination_cidr_block
            else ALLOWED_ROUTE_TARGET_GATEWAYS
        )
        new_target_key = get_common_key_in_dict(new_route, allowed_targets)
        if new_target_key is None or destination_cidr_key is None:
            result["result"] = False
            result["comment"] = (
                "There should be at least one valid destination cidr block and one target",
            )
            return result
        new_target_value = new_route.get(new_target_key)
        # Destination cidr is unique for a route. After retrieving destination cidr check in old routes
        # if the new route cidr block is not present in old routes we associate it to route table
        # if the destination cidr is already present in old routes we need to check if there is any changes in target
        # The allowed targets can be any one of Internet gateway, Virtual private gateway and many more.
        # If the target key is modified or value is modified we need to delete old route and create
        # a new route with updated values
        # example the target might be changed from an Internet gateway to Transit gateway the key changes
        # So we need to delete Internet gateway route and create Transit gateway route
        if destination_cidr_block not in old_routes_map:
            routes_to_add.append(
                {
                    destination_cidr_key: destination_cidr_block,
                    new_target_key: new_target_value,
                }
            )
        else:
            old_route = old_routes_map.get(destination_cidr_block)
            old_target_key = get_common_key_in_dict(
                old_route, ALLOWED_ROUTE_TARGET_GATEWAYS
            )
            old_target_value = old_route.get(old_target_key)
            if old_target_key != new_target_key or old_target_value != new_target_value:
                routes_to_replace.append(
                    {
                        destination_cidr_key: destination_cidr_block,
                        new_target_key: new_target_value,
                    }
                )
            del old_routes_map[destination_cidr_block]

    # delete the remaining routes which are still present in old routes and not in new routes
    for old_route in old_routes_map.values():
        destination_cidr_key = get_common_key_in_dict(
            old_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = old_route.get(destination_cidr_key)
        if old_route.get("Origin") != "CreateRouteTable":
            routes_to_delete.append({destination_cidr_key: destination_cidr_block})

    result["routes_to_add"] = routes_to_add
    result["routes_to_delete"] = routes_to_delete
    result["routes_to_replace"] = routes_to_replace
    return result


def get_common_key_in_dict(dict1: Dict, target_keys_list: Set):
    common_keys = set(dict1.keys()).intersection(target_keys_list)
    return common_keys.pop() if common_keys else None


def get_route_table_association_by_id(
    hub,
    resources: List,
    gateway_id: str = None,
    subnet_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    result = {"result": False, "resource_id": resource_id}
    if resource_id:
        key = "RouteTableAssociationId"
        value = resource_id
    elif subnet_id:
        key = "SubnetId"
        value = subnet_id
    else:
        key = "GatewayId"
        value = gateway_id
    if resources:
        for resource in resources:
            if key in resource and resource.get(key) == value:
                # Do not return ones that are in disassociated or disassociating state
                if resource["AssociationState"]["State"] == "associated":
                    result["ret"] = resource
                    result["result"] = True
                    break
    return result
